;;; vc-git-handler.el --- File Name Handler for revisions of Git versioned files  -*- lexical-binding:t -*-

;; Copyright (C) 2017 Free Software Foundation, Inc.

;; Author: Michael Albinus <michael.albinus@gmx.de>
;; Keywords: vc tools
;; Package: vc

;; This file is part of GNU Emacs.

;; GNU Emacs is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; This package provides file name handlers specific for Git.

;; A revision looks like "@@/master/ef7a18a071" or "@@/master/HEAD".
;; Branches are subdirectories, a revision in a branch might be
;; "@@/emacs-25/3a34412caa" or "@@/emacs-25/HEAD", and a label is a
;; symlink like "@@/emacs-25.2".  The element "HEAD" is always a
;; symlink to either the head of a branch, or to the current branch on
;; top level.

;;; Code:

(require 'vc-git)

;; New handlers should be added here.
(defconst vc-git-file-name-handler-alist
  '((file-attributes . vc-git-handle-file-attributes)
    (file-name-all-completions . vc-git-handle-file-name-all-completions)
    ;; TODO: Wouldn't it be better, to have `copy-file' here?
    (file-local-copy . vc-git-handle-file-local-copy))
  "Alist of Git specific handler functions.
Operations not mentioned here will be handled by vc-handler.el or
the default file name functions.")

;; Mark `operations' the handler is responsible for.
(put 'vc-file-name-handler 'operations
     (cl-union
      (get 'vc-file-name-handler 'operations)
      (mapcar 'car vc-git-file-name-handler-alist)))

;; Internal variables and functions.

(defvar vc-git-handler-branches nil
  "Cached local branch names.
The car of the list is the current branch.")

(defun vc-git-handler-branches ()
  "Return a list of local branches.
The car of the list is the current branch."
  (setq vc-git-handler-branches
	(vc-git-branches)))

(defvar vc-git-handler-tags nil
  "Cached tag names.")

(defun vc-git-handler-tags ()
  "Return a list of all tags."
  (setq vc-git-handler-tags
	(or vc-git-handler-tags
	    (split-string
	     (vc-git--run-command-string nil "tag") nil 'omit-nulls))))

(defvar vc-git-handler-heads nil
  "Cached alist of (OBJECT SHA1) tupels.
OBJECT is a branch name, a tag name, or \"HEAD\".")

(defun vc-git-handler-heads ()
  "Return an alist (OBJECT SHA1) tupels.
OBJECT is a branch name, a tag name, or \"HEAD\"."
  (setq vc-git-handler-heads
	(or vc-git-handler-heads
	    (mapcar
	     (lambda (x)
	       (list
		(replace-regexp-in-string
		 "refs/\\(tags\\|heads\\)/" "" (cadr x))
		(car x)))
	     (mapcar
	      ;; Hash object.
	      'split-string
	      ;; Lines.
	      (split-string
	       (vc-git--run-command-string
		nil "show-ref" "--heads" "--tags" "--head" "--abbrev")
	       "[\f\n]+" 'omit-nulls))))))

(defun vc-git-handler-head (object)
  "Return SHA1 of OBJECT.
OBJECT is a branch name, a tag name, or \"HEAD\"."
  (cadr (assoc object (vc-git-handler-heads))))

(defvar vc-git-handler-file-attributes (make-hash-table :test 'equal)
  "Cached file attributes.
It is a hash, the key is the revisioned file name, and the value
is the result of `file-attributes'.")

;; TODO: We shall add also functions to expire the caches.  Best would
;; be file notification, which watches respectice git files (indexes).

(defun vc-git-handler-object-exists-for-file-p (object filename)
  "Check, whether OBJECT (branch or tag) exists for FILE."
  ;; This is a sledge-hammer approach.  There must be something more
  ;; efficient.  For the time being, we simply return t.
  ;; (not
  ;;  (zerop
  ;;   (length
  ;;    (vc-git--run-command-string
  ;;     (vc-handler-file-name-part filename)
  ;;     "log" "--max-count=1" "--oneline" object "--")))))
  t)

;; The handlers.

(defun vc-git-handler-file-attributes-of-head (filename &optional id-format)
  "Like `file-attributes' for HEAD."
  (setq filename (expand-file-name filename))
  (let* ((file-name (vc-handler-file-name-part filename))
	 (revision (vc-handler-file-revision-name filename))
	 attr)
    ;; Revision is @@/branch/name/HEAD.
    (string-match "\\`@@\\(?:/\\(.*\\)\\)?/HEAD\\'" revision)
    (setq revision (match-string 1 revision)
	  attr
	  (file-attributes
	   (concat file-name "@@/" (vc-git-handler-head (or revision "HEAD")))
	   id-format))
    ;; Modify symlink.
    (if (zerop (length revision))
	(setcar attr (car (vc-git-handler-branches))) ;; Current branch.
      (setcar attr (vc-git-handler-head revision))) ;; Head of branch.
    (aset (nth 8 attr) 0 ?l)
    attr))

(defun vc-git-handle-file-attributes (filename &optional id-format)
  "Like `file-attributes' for revisioned files."
  (let ((cache-key (concat filename "@@" (symbol-name (or id-format 'integer))))
	attr)
    (cond
     ;; Cached value.
     ((setq attr (gethash cache-key vc-git-handler-file-attributes)))

     ;; Determine HEAD.
     ((string-equal (file-name-nondirectory filename) "HEAD")
      (setq attr (vc-git-handler-file-attributes-of-head filename id-format)))

     (t
      (setq filename (expand-file-name filename))
      (let* ((default-directory temporary-file-directory) ;; Avoid recursion.
             (file-name (vc-handler-file-name-part filename))
             (root (vc-git-root file-name))
             (default-directory (expand-file-name root))
	     (revision (vc-handler-file-revision-name filename))
	     git-log hash time author)
	(setq attr (file-attributes file-name id-format))
	;; Determine revision.
	(string-match "\\`@@/\\(.+\\)\\'" revision)
	(when (and (setq revision (match-string 1 revision))
		   ;; It could be branch/name/nnnnnnnnnn.
		   (file-name-directory revision)
		   (member
		    (directory-file-name (file-name-directory revision))
		    (vc-git-handler-branches)))
	  (setq revision (file-name-nondirectory revision)
		revision (unless (zerop (length revision)) revision)))

	;; Determine hash, commit time and commit author.
	(ignore-errors
	  (when (and (setq git-log
			   (vc-git--run-command-string
			    (unless (member revision (vc-git-handler-tags))
			      file-name)
			    "log" "--no-color" "--format=%h %at %an"
			    "--max-count=1" revision "--"))
		     (string-match
		      (concat
		       "\\`\\([[:alnum:]]+\\)[[:space:]]"
		       "\\([[:digit:]]+\\)[[:space:]]"
		       "\\(.+\\)\n?\\'")
		      git-log))
	    (setq hash (match-string 1 git-log)
		  time (string-to-number (match-string 2 git-log))
		  author (match-string 3 git-log))))

	;; Modify directory indicator.
	(when (or (null revision) (member revision (vc-git-handler-branches)))
	  (setcar attr t)
	  (aset (nth 8 attr) 0 ?d))

	;; Modify symlink.
	(when (member revision (vc-git-handler-tags))
	  (setcar attr hash)
	  (aset (nth 8 attr) 0 ?l))

	;; Modify uid and gid string.
	(when (and author (eq id-format 'string))
	  (setcar (nthcdr 2 attr) author)
	  (setcar (nthcdr 3 attr) "UNKNOWN"))

	;; Modify last access time, last modification time, and last
	;; status change time.
	(when time
	  (setcar
	   (nthcdr 4 attr) (list (floor time 65536) (floor (mod time 65536))))
	  (setcar
	   (nthcdr 5 attr) (list (floor time 65536) (floor (mod time 65536))))
	  (setcar
	   (nthcdr 6 attr) (list (floor time 65536) (floor (mod time 65536)))))

	;; Modify file size.
	(ignore-errors
	  (and revision
	       (setq git-log
		     (vc-git--run-command-string
		      nil "cat-file" "-s"
		      (format
		       "%s:%s" revision (file-relative-name file-name))))
	       (string-match "\\`\\([[:digit:]]+\\)\n?\\'" git-log)
	       (setcar
		(nthcdr 7 attr) (string-to-number (match-string 1 git-log)))))

	;; Modify mode string.  Remove write bit, and add execute bit
	;; for directories.
	(aset (nth 8 attr) 2 ?-)
	(aset (nth 8 attr) 5 ?-)
	(aset (nth 8 attr) 8 ?-)
	(when (char-equal (aref (nth 8 attr) 0) ?d)
	  (when (char-equal (aref (nth 8 attr) 1) ?r)
	    (aset (nth 8 attr) 3 ?x))
	  (when (char-equal (aref (nth 8 attr) 4) ?r)
	    (aset (nth 8 attr) 6 ?x))
	  (when (char-equal (aref (nth 8 attr) 7) ?r)
	    (aset (nth 8 attr) 9 ?x))))))

    ;; TODO: we need also to modify inode, device-number.

    ;; Result.
    (puthash cache-key attr vc-git-handler-file-attributes)))

;; This function should return "foo/" for directories and "bar" for files.
(defun vc-git-handle-file-name-all-completions (filename directory)
  "Like `file-name-all-completions' for revisioned files."
  (let* ((file-name (vc-handler-file-name-part directory))
	 (branch (vc-handler-file-revision-name directory))
	 (default-directory (unhandled-file-name-directory file-name))
	 base all-revisions all-tags all-branches)
    (unless (file-directory-p default-directory)
      (setq default-directory
	    (file-name-directory (directory-file-name default-directory))))
    ;; Read branch specific revisions.
    ;; TODO: This yields all revisions reachable from the branch head.
    ;; It might be better to return only revisions starting when the
    ;; branch was created, but I don't know how to determine this.
    ;; "git merge-base --fork-point <branch>" sounds like a good
    ;; candidate, but it doesn't work as expected.
    (string-match "\\`@@/\\(.+\\)\\'" branch)
    (when (and (setq branch (match-string 1 branch))
	       (setq branch (directory-file-name branch)))
      (ignore-errors
	(with-temp-buffer
	  (and
	   (vc-git-command
	    (current-buffer) nil file-name
	    "log" "--no-color" "--format=%h" branch "--")
	   (goto-char (point-min))
	   (while (< (point) (point-max))
	     (push
	      (buffer-substring-no-properties (point) (line-end-position))
	      all-revisions)
	     (forward-line 1))))))
    ;; Every branch has a virtual HEAD.
    (setq all-revisions (cons "HEAD" all-revisions))

    ;; Read tags.
    (setq all-tags
	  (mapcar
	   (lambda (x)
	     (and
	      (if branch
		  ;; Mention only tags belonging to branch.
		  (member (vc-git-handler-head x) all-revisions)
		;; All existing tags for that file.
		(vc-git-handler-object-exists-for-file-p x file-name))
	      x))
	   (vc-git-handler-tags)))

    ;; Read branches in top level for that file.  Add trailing "/".
    (unless branch
      (setq all-branches
	    (mapcar
	     (lambda (x)
	       (and (vc-git-handler-object-exists-for-file-p x file-name)
		    (file-name-as-directory x)))
	     (vc-git-handler-branches))))

    ;; Result.
    (all-completions
     filename (delq nil (append all-revisions all-tags all-branches)))))

(defun vc-git-handle-file-local-copy (filename)
  "Like `file-local-copy' for revisioned files."
  (setq filename (expand-file-name filename))
  (let* ((default-directory temporary-file-directory) ;; Avoid recursion.
         (file-name (vc-handler-file-name-part filename))
         (root (vc-git-root file-name))
         (default-directory (expand-file-name root))
	 (revision (vc-handler-file-revision-name filename))
	 (result
	  (make-temp-file "vc-" nil (file-name-extension file-name 'period))))
    ;; Determine revision.
    (string-match "\\`@@/\\(.+\\)\\'" revision)
    (when (setq revision (match-string 1 revision))
      (setq revision (file-name-nondirectory revision))
      (with-temp-buffer
        (and
	 (vc-git-command
	  (current-buffer) nil nil
	  "show" (format "%s:%s" revision (file-relative-name file-name)))
	 (write-region nil nil result)))
      ;; Set attributes.
      (set-file-times
       result (file-attribute-modification-time (file-attributes filename)))
      (set-file-modes result (file-modes filename))
      result)))

;; Debug.
(dolist (elt (all-completions "vc-git-" obarray 'functionp))
  (trace-function-background (intern elt)))

(provide 'vc-git-handler)

;;; vc-git-handler.el ends here

;; Local Variables:
;; mode: Emacs-Lisp
;; coding: utf-8
;; End:
